/*
 * ProGuard -- shrinking, optimization, obfuscation, and preverification
 *             of Java bytecode.
 *
 * Copyright (c) 2002-2017 Eric Lafortune @ GuardSquare
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package proguard.classfile.editor;

import proguard.classfile.*;
import proguard.classfile.attribute.visitor.AttributeVisitor;
import proguard.classfile.constant.Constant;
import proguard.classfile.instruction.*;
import proguard.classfile.util.*;
import proguard.optimize.peephole.InstructionSequenceReplacer;

import java.util.*;

/**
 * This utility class allows to construct sequences of instructions and
 * constants.
 *
 * @author Eric Lafortune
 */
public class InstructionSequenceBuilder
extends      SimplifiedVisitor
implements   AttributeVisitor
{
    protected final ProgramClass       dummyClass   = new ProgramClass(0, 0, new Constant[0], 0, 0, 0);
    protected final List<Instruction>  instructions = new ArrayList<Instruction>(256);
    protected final ConstantPoolEditor constantPoolEditor;


    /**
     * Creates a new InstructionSequenceBuilder.
     */
    public InstructionSequenceBuilder()
    {
        this(null, null);
    }


    /**
     * Creates a new InstructionSequenceBuilder that automatically initializes
     * class references and class member references in new constants.
     * @param programClassPool the program class pool from which new constants
     *                         can be initialized.
     * @param libraryClassPool the library class pool from which new constants
     *                         can be initialized.
     */
    public InstructionSequenceBuilder(ClassPool programClassPool,
                                      ClassPool libraryClassPool)
    {
        constantPoolEditor = new ConstantPoolEditor(dummyClass,
                                                    programClassPool,
                                                    libraryClassPool);
    }


    /**
     * Returns the ConstantPoolEditor used by this builder to
     * create constants.
     *
     * @return the ConstantPoolEditor used by this builder to
     * create constants.
     */
    public ConstantPoolEditor getConstantPoolEditor()
    {
        return constantPoolEditor;
    }


    /**
     * Short for {@link #appendInstruction(Instruction)}.
     *
     * @see InstructionSequenceReplacer#label()
     */
    public InstructionSequenceBuilder label(Instruction instruction)
    {
        return appendInstruction(instruction);
    }


    /**
     * Short for {@link #appendInstruction(Instruction)}.
     *
     * @see InstructionSequenceReplacer#catch_(int,int,int)
     */
    public InstructionSequenceBuilder catch_(Instruction instruction)
    {
        return appendInstruction(instruction);
    }


    /**
     * Appends the given instruction.
     * @param instruction the instruction to be appended.
     * @return the builder itself.
     */
    public InstructionSequenceBuilder appendInstruction(Instruction instruction)
    {
        instructions.add(instruction);
        return this;
    }


    /**
     * Appends the given instructions.
     * @param instructionArray the instructions to be appended.
     * @return the builder itself.
     */
    public InstructionSequenceBuilder appendInstructions(Instruction... instructionArray)
    {
        Collections.addAll(instructions, instructionArray);
        return this;
    }


    /**
     * Short for {@link #instructions()}.
     */
    public Instruction[] __()
    {
        return instructions();
    }


    /**
     * Returns the accumulated sequence of instructions
     * and resets the sequence in the builder.
     */
    public Instruction[] instructions()
    {
        Instruction[] instructions_array = new Instruction[instructions.size()];
        instructions.toArray(instructions_array);

        instructions.clear();

        return instructions_array;
    }


    /**
     * Returns the accumulated set of constants
     * and resets the set in the builder.
     */
    public Constant[] constants()
    {
        Constant[] constantPool = new Constant[dummyClass.u2constantPoolCount];
        System.arraycopy(dummyClass.constantPool,
                         0,
                         constantPool,
                         0,
                         dummyClass.u2constantPoolCount);

        dummyClass.u2constantPoolCount = 0;

        return constantPool;
    }


    // Methods corresponding to the bytecode opcodes.

    public InstructionSequenceBuilder nop()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_NOP));
        return this;
    }

    public InstructionSequenceBuilder aconst_null()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ACONST_NULL));
        return this;
    }

    public InstructionSequenceBuilder iconst(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_0, constant));
        return this;
    }

    public InstructionSequenceBuilder iconst_m1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_M1));
        return this;
    }

    public InstructionSequenceBuilder iconst_0()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_0));
        return this;
    }

    public InstructionSequenceBuilder iconst_1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_1));
        return this;
    }

    public InstructionSequenceBuilder iconst_2()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_2));
        return this;
    }

    public InstructionSequenceBuilder iconst_3()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_3));
        return this;
    }

    public InstructionSequenceBuilder iconst_4()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_4));
        return this;
    }

    public InstructionSequenceBuilder iconst_5()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ICONST_5));
        return this;
    }

    public InstructionSequenceBuilder lconst(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LCONST_0, constant));
        return this;
    }

    public InstructionSequenceBuilder lconst_0()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LCONST_0));
        return this;
    }

    public InstructionSequenceBuilder lconst_1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LCONST_1));
        return this;
    }

    public InstructionSequenceBuilder fconst(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FCONST_0, constant));
        return this;
    }

    public InstructionSequenceBuilder fconst_0()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FCONST_0));
        return this;
    }

    public InstructionSequenceBuilder fconst_1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FCONST_1));
        return this;
    }

    public InstructionSequenceBuilder fconst_2()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FCONST_2));
        return this;
    }

    public InstructionSequenceBuilder dconst(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DCONST_0, constant));
        return this;
    }

    public InstructionSequenceBuilder dconst_0()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DCONST_0));
        return this;
    }

    public InstructionSequenceBuilder dconst_1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DCONST_1));
        return this;
    }

    public InstructionSequenceBuilder bipush(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_BIPUSH, constant));
        return this;
    }

    public InstructionSequenceBuilder sipush(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_SIPUSH, constant));
        return this;
    }

    public InstructionSequenceBuilder ldc(int value)
    {
        return ldc_(constantPoolEditor.addIntegerConstant(value));
    }

    public InstructionSequenceBuilder ldc(float value)
    {
        return ldc_(constantPoolEditor.addFloatConstant(value));
    }

    public InstructionSequenceBuilder ldc(String string)
    {
        return ldc(string, null, null);
    }

    public InstructionSequenceBuilder ldc(String string, Clazz referencedClass, Method referencedMember)
    {
        return ldc_(constantPoolEditor.addStringConstant(string, referencedClass, referencedMember));
    }

    public InstructionSequenceBuilder ldc(String className, Clazz referencedClass)
    {
        return ldc_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder ldc_(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_LDC, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder ldc_w(int value)
    {
        return ldc_w_(constantPoolEditor.addIntegerConstant(value));
    }

    public InstructionSequenceBuilder ldc_w(float value)
    {
        return ldc_w_(constantPoolEditor.addFloatConstant(value));
    }

    public InstructionSequenceBuilder ldc_w(String string)
    {
        return ldc_w(string, null, null);
    }

    public InstructionSequenceBuilder ldc_w(String string, Clazz referencedClass, Method referencedMember)
    {
        return ldc_w_(constantPoolEditor.addStringConstant(string, referencedClass, referencedMember));
    }

    public InstructionSequenceBuilder ldc_w(String className, Clazz referencedClass)
    {
        return ldc_w_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder ldc_w_(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_LDC_W, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder ldc2_w(long value)
    {
        return ldc2_w(constantPoolEditor.addLongConstant(value));
    }

    public InstructionSequenceBuilder ldc2_w(double value)
    {
        return ldc2_w(constantPoolEditor.addDoubleConstant(value));
    }

    public InstructionSequenceBuilder ldc2_w(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_LDC2_W, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder iload(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ILOAD, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder lload(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LLOAD, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder fload(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FLOAD, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder dload(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DLOAD, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder aload(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ALOAD, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder iload_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ILOAD_0));
        return this;
    }

    public InstructionSequenceBuilder iload_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ILOAD_1));
        return this;
    }

    public InstructionSequenceBuilder iload_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ILOAD_2));
        return this;
    }

    public InstructionSequenceBuilder iload_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ILOAD_3));
        return this;
    }

    public InstructionSequenceBuilder lload_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LLOAD_0));
        return this;
    }

    public InstructionSequenceBuilder lload_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LLOAD_1));
        return this;
    }

    public InstructionSequenceBuilder lload_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LLOAD_2));
        return this;
    }

    public InstructionSequenceBuilder lload_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LLOAD_3));
        return this;
    }

    public InstructionSequenceBuilder fload_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FLOAD_0));
        return this;
    }

    public InstructionSequenceBuilder fload_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FLOAD_1));
        return this;
    }

    public InstructionSequenceBuilder fload_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FLOAD_2));
        return this;
    }

    public InstructionSequenceBuilder fload_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FLOAD_3));
        return this;
    }

    public InstructionSequenceBuilder dload_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DLOAD_0));
        return this;
    }

    public InstructionSequenceBuilder dload_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DLOAD_1));
        return this;
    }

    public InstructionSequenceBuilder dload_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DLOAD_2));
        return this;
    }

    public InstructionSequenceBuilder dload_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DLOAD_3));
        return this;
    }

    public InstructionSequenceBuilder aload_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ALOAD_0));
        return this;
    }

    public InstructionSequenceBuilder aload_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ALOAD_1));
        return this;
    }

    public InstructionSequenceBuilder aload_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ALOAD_2));
        return this;
    }

    public InstructionSequenceBuilder aload_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ALOAD_3));
        return this;
    }

    public InstructionSequenceBuilder iaload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IALOAD));
        return this;
    }

    public InstructionSequenceBuilder laload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LALOAD));
        return this;
    }

    public InstructionSequenceBuilder faload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FALOAD));
        return this;
    }

    public InstructionSequenceBuilder daload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DALOAD));
        return this;
    }

    public InstructionSequenceBuilder aaload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_AALOAD));
        return this;
    }

    public InstructionSequenceBuilder baload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_BALOAD));
        return this;
    }

    public InstructionSequenceBuilder caload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_CALOAD));
        return this;
    }

    public InstructionSequenceBuilder saload()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_SALOAD));
        return this;
    }

    public InstructionSequenceBuilder istore(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ISTORE, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder lstore(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LSTORE, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder fstore(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FSTORE, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder dstore(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DSTORE, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder astore(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ASTORE, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder istore_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ISTORE_0));
        return this;
    }

    public InstructionSequenceBuilder istore_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ISTORE_1));
        return this;
    }

    public InstructionSequenceBuilder istore_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ISTORE_2));
        return this;
    }

    public InstructionSequenceBuilder istore_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ISTORE_3));
        return this;
    }

    public InstructionSequenceBuilder lstore_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LSTORE_0));
        return this;
    }

    public InstructionSequenceBuilder lstore_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LSTORE_1));
        return this;
    }

    public InstructionSequenceBuilder lstore_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LSTORE_2));
        return this;
    }

    public InstructionSequenceBuilder lstore_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_LSTORE_3));
        return this;
    }

    public InstructionSequenceBuilder fstore_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FSTORE_0));
        return this;
    }

    public InstructionSequenceBuilder fstore_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FSTORE_1));
        return this;
    }

    public InstructionSequenceBuilder fstore_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FSTORE_2));
        return this;
    }

    public InstructionSequenceBuilder fstore_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_FSTORE_3));
        return this;
    }

    public InstructionSequenceBuilder dstore_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DSTORE_0));
        return this;
    }

    public InstructionSequenceBuilder dstore_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DSTORE_1));
        return this;
    }

    public InstructionSequenceBuilder dstore_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DSTORE_2));
        return this;
    }

    public InstructionSequenceBuilder dstore_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_DSTORE_3));
        return this;
    }

    public InstructionSequenceBuilder astore_0()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ASTORE_0));
        return this;
    }

    public InstructionSequenceBuilder astore_1()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ASTORE_1));
        return this;
    }

    public InstructionSequenceBuilder astore_2()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ASTORE_2));
        return this;
    }

    public InstructionSequenceBuilder astore_3()
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_ASTORE_3));
        return this;
    }

    public InstructionSequenceBuilder iastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IASTORE));
        return this;
    }

    public InstructionSequenceBuilder lastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LASTORE));
        return this;
    }

    public InstructionSequenceBuilder fastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FASTORE));
        return this;
    }

    public InstructionSequenceBuilder dastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DASTORE));
        return this;
    }

    public InstructionSequenceBuilder aastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_AASTORE));
        return this;
    }

    public InstructionSequenceBuilder bastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_BASTORE));
        return this;
    }

    public InstructionSequenceBuilder castore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_CASTORE));
        return this;
    }

    public InstructionSequenceBuilder sastore()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_SASTORE));
        return this;
    }

    public InstructionSequenceBuilder pop()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_POP));
        return this;
    }

    public InstructionSequenceBuilder pop2()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_POP2));
        return this;
    }

    public InstructionSequenceBuilder dup()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DUP));
        return this;
    }

    public InstructionSequenceBuilder dup_x1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DUP_X1));
        return this;
    }

    public InstructionSequenceBuilder dup_x2()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DUP_X2));
        return this;
    }

    public InstructionSequenceBuilder dup2()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DUP2));
        return this;
    }

    public InstructionSequenceBuilder dup2_x1()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DUP2_X1));
        return this;
    }

    public InstructionSequenceBuilder dup2_x2()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DUP2_X2));
        return this;
    }

    public InstructionSequenceBuilder swap()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_SWAP));
        return this;
    }

    public InstructionSequenceBuilder iadd()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IADD));
        return this;
    }

    public InstructionSequenceBuilder ladd()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LADD));
        return this;
    }

    public InstructionSequenceBuilder fadd()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FADD));
        return this;
    }

    public InstructionSequenceBuilder dadd()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DADD));
        return this;
    }

    public InstructionSequenceBuilder isub()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ISUB));
        return this;
    }

    public InstructionSequenceBuilder lsub()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LSUB));
        return this;
    }

    public InstructionSequenceBuilder fsub()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FSUB));
        return this;
    }

    public InstructionSequenceBuilder dsub()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DSUB));
        return this;
    }

    public InstructionSequenceBuilder imul()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IMUL));
        return this;
    }

    public InstructionSequenceBuilder lmul()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LMUL));
        return this;
    }

    public InstructionSequenceBuilder fmul()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FMUL));
        return this;
    }

    public InstructionSequenceBuilder dmul()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DMUL));
        return this;
    }

    public InstructionSequenceBuilder idiv()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IDIV));
        return this;
    }

    public InstructionSequenceBuilder ldiv()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LDIV));
        return this;
    }

    public InstructionSequenceBuilder fdiv()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FDIV));
        return this;
    }

    public InstructionSequenceBuilder ddiv()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DDIV));
        return this;
    }

    public InstructionSequenceBuilder irem()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IREM));
        return this;
    }

    public InstructionSequenceBuilder lrem()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LREM));
        return this;
    }

    public InstructionSequenceBuilder frem()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FREM));
        return this;
    }

    public InstructionSequenceBuilder drem()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DREM));
        return this;
    }

    public InstructionSequenceBuilder ineg()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_INEG));
        return this;
    }

    public InstructionSequenceBuilder lneg()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LNEG));
        return this;
    }

    public InstructionSequenceBuilder fneg()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FNEG));
        return this;
    }

    public InstructionSequenceBuilder dneg()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DNEG));
        return this;
    }

    public InstructionSequenceBuilder ishl()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ISHL));
        return this;
    }

    public InstructionSequenceBuilder lshl()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LSHL));
        return this;
    }

    public InstructionSequenceBuilder ishr()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ISHR));
        return this;
    }

    public InstructionSequenceBuilder lshr()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LSHR));
        return this;
    }

    public InstructionSequenceBuilder iushr()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IUSHR));
        return this;
    }

    public InstructionSequenceBuilder lushr()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LUSHR));
        return this;
    }

    public InstructionSequenceBuilder iand()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IAND));
        return this;
    }

    public InstructionSequenceBuilder land()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LAND));
        return this;
    }

    public InstructionSequenceBuilder ior()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IOR));
        return this;
    }

    public InstructionSequenceBuilder lor()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LOR));
        return this;
    }

    public InstructionSequenceBuilder ixor()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IXOR));
        return this;
    }

    public InstructionSequenceBuilder lxor()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LXOR));
        return this;
    }

    public InstructionSequenceBuilder iinc(int variableIndex,
                                           int constant)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_IINC, variableIndex, constant));
        return this;
    }

    public InstructionSequenceBuilder i2l()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_I2L));
        return this;
    }

    public InstructionSequenceBuilder i2f()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_I2F));
        return this;
    }

    public InstructionSequenceBuilder i2d()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_I2D));
        return this;
    }

    public InstructionSequenceBuilder l2i()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_L2I));
        return this;
    }

    public InstructionSequenceBuilder l2f()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_L2F));
        return this;
    }

    public InstructionSequenceBuilder l2d()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_L2D));
        return this;
    }

    public InstructionSequenceBuilder f2i()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_F2I));
        return this;
    }

    public InstructionSequenceBuilder f2l()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_F2L));
        return this;
    }

    public InstructionSequenceBuilder f2d()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_F2D));
        return this;
    }

    public InstructionSequenceBuilder d2i()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_D2I));
        return this;
    }

    public InstructionSequenceBuilder d2l()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_D2L));
        return this;
    }

    public InstructionSequenceBuilder d2f()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_D2F));
        return this;
    }

    public InstructionSequenceBuilder i2b()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_I2B));
        return this;
    }

    public InstructionSequenceBuilder i2c()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_I2C));
        return this;
    }

    public InstructionSequenceBuilder i2s()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_I2S));
        return this;
    }

    public InstructionSequenceBuilder lcmp()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LCMP));
        return this;
    }

    public InstructionSequenceBuilder fcmpl()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FCMPL));
        return this;
    }

    public InstructionSequenceBuilder fcmpg()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FCMPG));
        return this;
    }

    public InstructionSequenceBuilder dcmpl()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DCMPL));
        return this;
    }

    public InstructionSequenceBuilder dcmpg()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DCMPG));
        return this;
    }

    public InstructionSequenceBuilder ifeq(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFEQ, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifne(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFNE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder iflt(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFLT, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifge(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFGE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifgt(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFGT, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifle(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFLE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ificmpeq(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFICMPEQ, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ificmpne(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFICMPNE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ificmplt(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFICMPLT, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ificmpge(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFICMPGE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ificmpgt(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFICMPGT, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ificmple(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFICMPLE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifacmpeq(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFACMPEQ, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifacmpne(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFACMPNE, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder goto_(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_GOTO, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder jsr(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_JSR, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ret(int variableIndex)
    {
        instructions.add(new VariableInstruction(InstructionConstants.OP_RET, variableIndex));
        return this;
    }

    public InstructionSequenceBuilder tableswitch(int   defaultOffset,
                                                    int   lowCase,
                                                    int   highCase,
                                                    int[] jumpOffsets)
    {
        instructions.add(new TableSwitchInstruction(InstructionConstants.OP_TABLESWITCH,
                                                    defaultOffset,
                                                    lowCase,
                                                    highCase,
                                                    jumpOffsets));
        return this;
    }

    public InstructionSequenceBuilder lookupswitch(int  defaultOffset,
                                                    int[] cases,
                                                    int[] jumpOffsets)
    {
        instructions.add(new LookUpSwitchInstruction(InstructionConstants.OP_LOOKUPSWITCH,
                                                     defaultOffset,
                                                     cases,
                                                     jumpOffsets));
        return this;
    }

    public InstructionSequenceBuilder ireturn()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_IRETURN));
        return this;
    }

    public InstructionSequenceBuilder lreturn()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_LRETURN));
        return this;
    }

    public InstructionSequenceBuilder freturn()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_FRETURN));
        return this;
    }

    public InstructionSequenceBuilder dreturn()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_DRETURN));
        return this;
    }

    public InstructionSequenceBuilder areturn()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ARETURN));
        return this;
    }

    public InstructionSequenceBuilder return_()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_RETURN));
        return this;
    }

    public InstructionSequenceBuilder getstatic(String className,
                                                String name,
                                                String descriptor)
    {
        return getstatic(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder getstatic(String className,
                                                String name,
                                                String descriptor,
                                                Clazz  referencedClass,
                                                Member referencedMember)
    {
        return getstatic(constantPoolEditor.addFieldrefConstant(className,
                                                                name,
                                                                descriptor,
                                                                referencedClass,
                                                                referencedMember));
    }

    public InstructionSequenceBuilder getstatic(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_GETSTATIC, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder putstatic(String className,
                                                String name,
                                                String descriptor)
    {
        return putstatic(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder putstatic(String className,
                                                String name,
                                                String descriptor,
                                                Clazz  referencedClass,
                                                Member referencedMember)
    {
        return putstatic(constantPoolEditor.addFieldrefConstant(className,
                                                                name,
                                                                descriptor,
                                                                referencedClass,
                                                                referencedMember));
    }

    public InstructionSequenceBuilder putstatic(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_PUTSTATIC, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder getfield(String className,
                                               String name,
                                               String descriptor)
    {
        return getfield(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder getfield(String className,
                                               String name,
                                               String descriptor,
                                               Clazz  referencedClass,
                                               Member referencedMember)
    {
        return getfield(constantPoolEditor.addFieldrefConstant(className,
                                                               name,
                                                               descriptor,
                                                               referencedClass,
                                                               referencedMember));
    }

    public InstructionSequenceBuilder getfield(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_GETFIELD, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder putfield(String className,
                                               String name,
                                               String descriptor)
    {
        return putfield(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder putfield(String className,
                                               String name,
                                               String descriptor,
                                               Clazz  referencedClass,
                                               Member referencedMember)
    {
        return putfield(constantPoolEditor.addFieldrefConstant(className,
                                                               name,
                                                               descriptor,
                                                               referencedClass,
                                                               referencedMember));
    }

    public InstructionSequenceBuilder putfield(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_PUTFIELD, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder invokevirtual(String className,
                                                    String name,
                                                    String descriptor)
    {
        return invokevirtual(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder invokevirtual(int    classIndex,
                                                    String name,
                                                    String descriptor)
    {
        return invokevirtual(constantPoolEditor.addMethodrefConstant(classIndex,
                                                                     name,
                                                                     descriptor,
                                                                     null,
                                                                     null));
    }

    public InstructionSequenceBuilder invokevirtual(String className,
                                                    String name,
                                                    String descriptor,
                                                    Clazz  referencedClass,
                                                    Member referencedMember)
    {
        return invokevirtual(constantPoolEditor.addMethodrefConstant(className,
                                                                     name,
                                                                     descriptor,
                                                                     referencedClass,
                                                                     referencedMember));
    }

    public InstructionSequenceBuilder invokevirtual(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_INVOKEVIRTUAL, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder invokespecial(String className,
                                                    String name,
                                                    String descriptor)
    {
        return invokespecial(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder invokespecial(String className,
                                                    String name,
                                                    String descriptor,
                                                    Clazz  referencedClass,
                                                    Member referencedMember)
    {
        return invokespecial(constantPoolEditor.addMethodrefConstant(className,
                                                                     name,
                                                                     descriptor,
                                                                     referencedClass,
                                                                     referencedMember));
    }

    public InstructionSequenceBuilder invokespecial(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_INVOKESPECIAL, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder invokestatic(String className,
                                                   String name,
                                                   String descriptor)
    {
        return invokestatic(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder invokestatic(String className,
                                                   String name,
                                                   String descriptor,
                                                   Clazz  referencedClass,
                                                   Member referencedMember)
    {
        return invokestatic(constantPoolEditor.addMethodrefConstant(className,
                                                                    name,
                                                                    descriptor,
                                                                    referencedClass,
                                                                    referencedMember));
    }

    public InstructionSequenceBuilder invokestatic(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_INVOKESTATIC, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder invokeinterface(String className,
                                                      String name,
                                                      String descriptor)
    {
        return invokeinterface(className, name, descriptor, null, null);
    }

    public InstructionSequenceBuilder invokeinterface(String className,
                                                      String name,
                                                      String descriptor,
                                                      Clazz  referencedClass,
                                                      Member referencedMember)
    {
        int invokeinterfaceConstant =
            (ClassUtil.internalMethodParameterSize(descriptor, false)) << 8;

        return invokeinterface(constantPoolEditor.addInterfaceMethodrefConstant(className,
                                                                                name,
                                                                                descriptor,
                                                                                referencedClass,
                                                                                referencedMember),
                               invokeinterfaceConstant);
    }

    public InstructionSequenceBuilder invokeinterface(int constantIndex, int constant)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_INVOKEINTERFACE, constantIndex, constant));
        return this;
    }

    public InstructionSequenceBuilder invokedynamic(int    bootStrapMethodIndex,
                                                    String name,
                                                    String descriptor)
    {
        return invokedynamic(bootStrapMethodIndex, name, descriptor, null);
    }

    public InstructionSequenceBuilder invokedynamic(int     bootStrapMethodIndex,
                                                    String  name,
                                                    String  descriptor,
                                                    Clazz[] referencedClasses)
    {
        return invokedynamic(constantPoolEditor.addInvokeDynamicConstant(bootStrapMethodIndex,
                                                                         name,
                                                                         descriptor,
                                                                         referencedClasses));
    }

    public InstructionSequenceBuilder invokedynamic(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_INVOKEDYNAMIC, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder new_(String className)
    {
        return new_(className, null);
    }

    public InstructionSequenceBuilder new_(String className, Clazz referencedClass)
    {
        return new_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder new_(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_NEW, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder newarray(int constant)
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_NEWARRAY, constant));
        return this;
    }

    public InstructionSequenceBuilder anewarray(String className, Clazz referencedClass)
    {
        return anewarray(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder anewarray(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_ANEWARRAY, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder arraylength()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ARRAYLENGTH));
        return this;
    }

    public InstructionSequenceBuilder athrow()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_ATHROW));
        return this;
    }

    public InstructionSequenceBuilder checkcast(String className, Clazz referencedClass)
    {
        return checkcast(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder checkcast(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_CHECKCAST, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder instanceof_(String className, Clazz referencedClass)
    {
        return instanceof_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder instanceof_(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_INSTANCEOF, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder monitorenter()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_MONITORENTER));
        return this;
    }

    public InstructionSequenceBuilder monitorexit()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_MONITOREXIT));
        return this;
    }

    public InstructionSequenceBuilder wide()
    {
        instructions.add(new SimpleInstruction(InstructionConstants.OP_WIDE));
        return this;
    }

    public InstructionSequenceBuilder multianewarray(String className, Clazz referencedClass)
    {
        return multianewarray(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public InstructionSequenceBuilder multianewarray(int constantIndex)
    {
        instructions.add(new ConstantInstruction(InstructionConstants.OP_MULTIANEWARRAY, constantIndex));
        return this;
    }

    public InstructionSequenceBuilder ifnull(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFNULL, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder ifnonnull(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_IFNONNULL, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder goto_w(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_GOTO_W, branchOffset));
        return this;
    }

    public InstructionSequenceBuilder jsr_w(int branchOffset)
    {
        instructions.add(new BranchInstruction(InstructionConstants.OP_JSR_W, branchOffset));
        return this;
    }


    public static void main(String[] args)
    {
        InstructionSequenceBuilder builder = new InstructionSequenceBuilder();

        Instruction[] instructions = builder
            .iconst_2()
            .istore_0()
            .iinc(0, 2)
            .iload_0()
            .ldc(12)
            .iadd()
            .putstatic("com/example/SomeClass", "someField", "I", null, null)
            .instructions();

        Constant[] constants = builder.constants();

        for (Instruction instruction : instructions)
        {
            System.out.println(instruction);
        }

        System.out.println();

        for (int index = 0; index < constants.length; index++)
        {
            System.out.println("#"+index+": " + constants[index]);
        }
    }
}
