/*
 * ProGuard -- shrinking, optimization, obfuscation, and preverification
 *             of Java bytecode.
 *
 * Copyright (c) 2002-2017 Eric Lafortune @ GuardSquare
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package proguard.classfile.util;

import proguard.classfile.*;
import proguard.classfile.attribute.*;
import proguard.classfile.attribute.visitor.*;
import proguard.classfile.constant.*;
import proguard.classfile.constant.visitor.*;
import proguard.classfile.editor.*;
import proguard.classfile.instruction.*;
import proguard.classfile.instruction.visitor.InstructionVisitor;
import proguard.classfile.visitor.*;
import proguard.optimize.info.MutableBoolean;


/**
 * This ClassVisitor will replace all PrimitiveArray constants
 * with Java byte-code compliant array store instructions.
 *
 * @author Thomas Neidhart
 */
public class PrimitiveArrayConstantReplacer
extends      SimplifiedVisitor
implements   ClassVisitor,
             AttributeVisitor,
             InstructionVisitor,
             ConstantVisitor
{
    private final CodeAttributeEditor codeAttributeEditor = new CodeAttributeEditor();
    private final MutableBoolean      classModified       = new MutableBoolean();

    private PrimitiveArrayConstant    referencedConstant;


    // Implementations for ClassVisitor.

    public void visitAnyClass(Clazz clazz) {}


    public void visitProgramClass(ProgramClass programClass)
    {
        ConstantCounter counter = new ConstantCounter();
        programClass.constantPoolEntriesAccept(
            new ConstantTagFilter(ClassConstants.CONSTANT_PrimitiveArray,
            counter));

        // Replace PrimitiveArray constants if the class has some.
        if (counter.getCount() > 0)
        {
            classModified.reset();

            programClass.methodsAccept(new AllAttributeVisitor(this));

            if (classModified.isSet())
            {
                // The class has been modified, let the ConstantPoolShrinker
                // remove the now unused PrimitiveArrayConstants.
                programClass.accept(new ConstantPoolShrinker());
            }
        }
    }


    // Implementations for AttributeVisitor.

    public void visitAnyAttribute(Clazz clazz, Attribute attribute) {}


    public void visitCodeAttribute(Clazz clazz, Method method, CodeAttribute codeAttribute)
    {
        codeAttributeEditor.reset(codeAttribute.u4codeLength);

        codeAttribute.instructionsAccept(clazz, method, this);

        if (codeAttributeEditor.isModified())
        {
            codeAttributeEditor.visitCodeAttribute(clazz, method, codeAttribute);
            classModified.set();
        }
    }


    // Implementations for InstructionVisitor.

    public void visitAnyInstruction(Clazz clazz, Method method, CodeAttribute codeAttribute, int offset, Instruction instruction) {}


    public void visitConstantInstruction(Clazz clazz, Method method, CodeAttribute codeAttribute, int offset, ConstantInstruction constantInstruction)
    {
        if (constantInstruction.opcode == InstructionConstants.OP_LDC ||
            constantInstruction.opcode == InstructionConstants.OP_LDC_W)
        {
            referencedConstant = null;
            clazz.constantPoolEntryAccept(constantInstruction.constantIndex, this);

            if (referencedConstant != null)
            {
                CompactInstructionListComposer composer =
                    new CompactInstructionListComposer((ProgramClass) clazz);

                int arrayLength = referencedConstant.getLength();
                Object values   = referencedConstant.getValues();

                char primitiveType = referencedConstant.getPrimitiveType();
                String elementType = String.valueOf(primitiveType);

                composer.ldc(arrayLength);
                composer.newarray(InstructionUtil.arrayTypeFromInternalType(primitiveType));
                for (int i = 0; i < arrayLength; i++)
                {
                    composer.dup();
                    composer.pushInt(i);
                    composer.pushPrimitive(getElementAt(values, i), primitiveType);
                    composer.storeToArray(elementType);
                }

                codeAttributeEditor.replaceInstruction(offset, composer.getInstructionsAsArray());
            }
        }
    }


    // Implementations for ConstantVisitor.

    public void visitAnyConstant(Clazz clazz, Constant constant) {}


    public void visitPrimitiveArrayConstant(Clazz clazz, PrimitiveArrayConstant primitiveArrayConstant)
    {
        referencedConstant = primitiveArrayConstant;
    }


    // Private utility methods.

    private Object getElementAt(Object values, int index)
    {
        // Note: use if-else constructs instead of ternary
        //       operator as the implicit object conversion
        //       leads to wrong byte code with Java 8 compiler.

        if (values instanceof boolean[])
        {
            return ((boolean[]) values)[index];
        }
        else if (values instanceof byte[])
        {
            return ((byte[]) values)[index];
        }
        else if (values instanceof char[])
        {
            return ((char[]) values)[index];
        }
        else if (values instanceof short[])
        {
            return ((short[]) values)[index];
        }
        else if (values instanceof int[])
        {
            return ((int[]) values)[index];
        }
        else if (values instanceof long[])
        {
            return ((long[]) values)[index];
        }
        else if (values instanceof float[])
        {
            return ((float[]) values)[index];
        }
        else if (values instanceof double[])
        {
            return ((double[]) values)[index];
        }
        else
        {
            throw new IllegalArgumentException("Invalid primitive array");
        }
    }

}
